﻿using UnityEngine;
using System.Collections;
using System.Collections.Generic;

public class LevelModel : MonoBehaviour {

	public enum TileType {Empty, Gravel, Stone, Hole, Lock, Door, Trap, Finish};

	public Camera mainCamera;
	public GameObject playerHolderPrefab;
	public GameObject gravelPrefab;
	public GameObject shadowGravelPrefab;
	public GameObject stonePrefab;
	public GameObject trapPrefab;
	public GameObject keyPrefab;
	public GameObject lockPrefab;
	public GameObject doorPrefab;
	public GameObject finishPrefab;
	public GameObject holePrefab;

	public GameObject levelBoard;
	public GameObject levelBoardModel;
	public GameObject topLevel;
	public GameObject bottomLevel;

	public int levelWidth;
	public int levelHeight;
	public bool playerOnTop;

	public int playerPosX;
	public int playerPosY;

	private TileType[,] topTiles;
	private TileType[,] bottomTiles;
	private List<Key> keys;
	private List<Door> doors;

	private GameObject player;

	private GameObject[,] topGravel;
	private GameObject[,] bottomGravel;
	

	public static LevelModel Instance {
		get {
			return instance;
		}
	}
	
	private static LevelModel instance;
	
	void Awake() {
		
		instance = this;
		keys = new List<Key> ();
		doors = new List<Door>();
	}

	void OnDestroy() {

		instance = null;
		PlayerPrefs.DeleteKey("level");
	}

	void Start () {
	
	}
	
	void Update () {

	}

	void CreateLevel() {

		string data = LevelData.LevelDataForCurrentLevel (ref levelWidth, ref levelHeight);

		if (data.Length != levelWidth * levelHeight * 2) {
			throw new System.Exception();
		}

		mainCamera.transform.position = new Vector3 (0, Mathf.Max(10, levelHeight) * 0.7f / Mathf.Tan (mainCamera.fieldOfView * Mathf.Deg2Rad / 2), 0);
		levelBoardModel.transform.localScale = new Vector3 (levelWidth, 0.5f, levelHeight);

		topTiles = new TileType[levelWidth, levelHeight];
		bottomTiles = new TileType[levelWidth, levelHeight];

		topGravel = new GameObject[levelWidth, levelHeight];
		bottomGravel = new GameObject[levelWidth, levelHeight];

		List<Lock> doorLocks = new List<Lock>();
		keys.Clear ();
		doors.Clear ();
		Destroy (player);
		player = null;
		playerOnTop = true;
		StopAllCoroutines ();
		levelBoard.GetComponent<BoardFlipControl> ().Reset ();
		Transform[] children;
		children = topLevel.GetComponentsInChildren<Transform>();
		foreach (Transform childTransform in children) {
			if (!topLevel.transform.IsChildOf(childTransform)) {
				Destroy(childTransform.gameObject);
			}
		}
		children = bottomLevel.GetComponentsInChildren<Transform>();
		foreach (Transform childTransform in children) {
			if (!bottomLevel.transform.IsChildOf(childTransform)) {
				Destroy(childTransform.gameObject);
			}
		}

		for (int i = 0; i < data.Length; i++) {
			int posX = i % levelWidth;
			int posY = levelHeight - 1 - i / levelWidth;
			bool forTopLevel = true;
			if (posY < 0) {
				posY = levelHeight + posY;
				forTopLevel = false;
			}

			Transform levelTransform = forTopLevel ? topLevel.transform : bottomLevel.transform;
			GameObject tileObject = null;
			TileType tileType = TileType.Empty;
			switch(data[i]) {
			case '-' : {
				tileObject = (GameObject)Instantiate(gravelPrefab);
				tileType = TileType.Gravel;
				break;
			}
			case 'X' : {
				tileObject = (GameObject)Instantiate(stonePrefab);
				tileType = TileType.Stone;
				break;
			}
			case 'T' : {
				tileObject = (GameObject)Instantiate(trapPrefab);
				tileType = TileType.Trap;
				break;
			}
			case 'O' : {
				tileObject = (GameObject)Instantiate(holePrefab);
				tileType = TileType.Hole;
				break;
			}
			case 'F' : {
				tileObject = (GameObject)Instantiate(finishPrefab);
				tileType = TileType.Finish;
				break;
			}
			case 'P' : {
				if (player == null) {
					playerPosX = posX;
					playerPosY = posY;
					player = (GameObject)Instantiate(playerHolderPrefab);
					player.transform.rotation = levelBoard.transform.rotation;
					player.transform.parent = levelBoard.transform;
					player.transform.localPosition = TileObjectPosFromPosXY(posX, posY, forTopLevel);
					player.GetComponent<PlayerControl>().levelBoard = levelBoard;
				}
				break;
			}
			case '1' : {
				tileObject = (GameObject)Instantiate(doorPrefab);
				tileType = TileType.Door;
				Door door = tileObject.GetComponent<Door>();
				door.doorIndex = 1;
				door.forTopLevel = forTopLevel;
				door.posX = posX;
				door.posY = posY;
				door.SetColor(Color.red);
				doors.Add(door);
				Key key = KeyForIndex(door.doorIndex);
				if (key) {
					key.door = door;
					door.doorKey = key;
				}
				break;
			}
			case '2' : {
				tileObject = (GameObject)Instantiate(doorPrefab);
				tileType = TileType.Door;
				Door door = tileObject.GetComponent<Door>();
				door.doorIndex = 2;
				door.forTopLevel = forTopLevel;
				door.posX = posX;
				door.posY = posY;
				door.SetColor(Color.green);
				doors.Add(door);
				Key key = KeyForIndex(door.doorIndex);
				if (key) {
					key.door = door;
					door.doorKey = key;
				}
				break;
			}
			case '3' : {
				tileObject = (GameObject)Instantiate(doorPrefab);
				tileType = TileType.Door;
				Door door = tileObject.GetComponent<Door>();
				door.doorIndex = 3;
				door.forTopLevel = forTopLevel;
				door.posX = posX;
				door.posY = posY;
				door.SetColor(Color.blue);
				doors.Add(door);
				Key key = KeyForIndex(door.doorIndex);
				if (key) {
					key.door = door;
					door.doorKey = key;
				}
				break;
			}
			case 'r' : {
				tileObject = (GameObject)Instantiate(keyPrefab);
				Key key = tileObject.GetComponent<Key>();
				key.keyIndex = 1;
				key.forTopLevel = forTopLevel;
				key.posX = posX;
				key.posY = posY;
				key.SetColor(Color.red);
				keys.Add(key);
				Door door = DoorForIndex(key.keyIndex);
				if (door) {
					key.door = door;
					door.doorKey = key;
				}
				Lock doorLock = DoorLockForIndex(key.keyIndex, doorLocks);
				if (doorLock) {
					key.doorLock = doorLock;
					doorLock.doorKey = key;
				}
				break;
			}
			case 'g' : {
				tileObject = (GameObject)Instantiate(keyPrefab);
				Key key = tileObject.GetComponent<Key>();
				key.keyIndex = 2;
				key.forTopLevel = forTopLevel;
				key.posX = posX;
				key.posY = posY;
				key.SetColor(Color.green);
				keys.Add(key);
				Door door = DoorForIndex(key.keyIndex);
				if (door) {
					key.door = door;
					door.doorKey = key;
				}
				Lock doorLock = DoorLockForIndex(key.keyIndex, doorLocks);
				if (doorLock) {
					key.doorLock = doorLock;
					doorLock.doorKey = key;
				}
				break;
			}
			case 'b' : {
				tileObject = (GameObject)Instantiate(keyPrefab);
				Key key = tileObject.GetComponent<Key>();
				key.keyIndex = 3;
				key.forTopLevel = forTopLevel;
				key.posX = posX;
				key.posY = posY;
				key.SetColor(Color.blue);
				keys.Add(key);
				Door door = DoorForIndex(key.keyIndex);
				if (door) {
					key.door = door;
					door.doorKey = key;
				}
				Lock doorLock = DoorLockForIndex(key.keyIndex, doorLocks);
				if (doorLock) {
					key.doorLock = doorLock;
					doorLock.doorKey = key;
				}
				break;
			}
			case 'R' : {
				tileObject = (GameObject)Instantiate(lockPrefab);
				tileType = TileType.Lock;
				Lock doorLock = tileObject.GetComponent<Lock>();
				doorLock.lockIndex = 1;
				doorLock.posX = posX;
				doorLock.posY = posY;
				doorLock.SetColor(Color.red);
				doorLocks.Add(doorLock);
				Key key = KeyForIndex(doorLock.lockIndex);
				if (key) {
					key.doorLock = doorLock;
					doorLock.doorKey = key;
				}
				break;
			}
			case 'G' : {
				tileObject = (GameObject)Instantiate(lockPrefab);
				tileType = TileType.Lock;
				Lock doorLock = tileObject.GetComponent<Lock>();
				doorLock.lockIndex = 2;
				doorLock.posX = posX;
				doorLock.posY = posY;
				doorLock.SetColor(Color.green);
				doorLocks.Add(doorLock);
				Key key = KeyForIndex(doorLock.lockIndex);
				if (key) {
					key.doorLock = doorLock;
					doorLock.doorKey = key;
				}
				break;
			}
			case 'B' : {
				tileObject = (GameObject)Instantiate(lockPrefab);
				tileType = TileType.Lock;
				Lock doorLock = tileObject.GetComponent<Lock>();
				doorLock.lockIndex = 3;
				doorLock.posX = posX;
				doorLock.posY = posY;
				doorLock.SetColor(Color.blue);
				doorLocks.Add(doorLock);
				Key key = KeyForIndex(doorLock.lockIndex);
				if (key) {
					key.doorLock = doorLock;
					doorLock.doorKey = key;
				}
				break;
			}
			default: {
				break;
			}
			}
			if (tileObject) {
				tileObject.transform.rotation = levelTransform.rotation;
				tileObject.transform.parent = levelTransform;
				tileObject.transform.localPosition = TileObjectPosFromPosXY(posX, posY, forTopLevel);
			}
			if (forTopLevel) {
				topTiles[posX, posY] = tileType;
				if (tileType == TileType.Gravel || tileType == TileType.Trap) {
					topGravel[posX, posY] = tileObject;
				}
			}
			else {
				bottomTiles[posX, posY] = tileType;
				if (tileType == TileType.Gravel || tileType == TileType.Trap) {
					bottomGravel[posX, posY] = tileObject;
				}
			}
		}

		for (int posY = 0; posY < levelHeight; posY++) {
			for (int posX = 0; posX < levelWidth; posX++) {
				if (topTiles[posX, posY] == TileType.Empty && bottomTiles[posX, posY] == TileType.Gravel) {
					GameObject tileObject = (GameObject)Instantiate(shadowGravelPrefab);
					tileObject.transform.rotation = topLevel.transform.rotation;
					tileObject.transform.parent = topLevel.transform;
					tileObject.transform.localPosition = TileObjectPosFromPosXY(posX, posY, true);
					topGravel[posX, posY] = tileObject;
				}
				else if (bottomTiles[posX, posY] == TileType.Empty && topTiles[posX, posY] == TileType.Gravel) {
					GameObject tileObject = (GameObject)Instantiate(shadowGravelPrefab);
					tileObject.transform.rotation = bottomLevel.transform.rotation;
					tileObject.transform.parent = bottomLevel.transform;
					tileObject.transform.localPosition = TileObjectPosFromPosXY(posX, posY, false);
					bottomGravel[posX, posY] = tileObject;
				}
			}
		}
	}

	Key KeyForIndex(int index) {

		foreach (Key key in keys) {
			if (key.keyIndex == index) {
				return key;
			}
		}
		return null;
	}

	Door DoorForIndex(int index) {
		
		foreach (Door door in doors) {
			if (door.doorIndex == index) {
				return door;
			}
		}
		return null;
	}

	Lock DoorLockForIndex(int index, List<Lock> doorLocks) {
		
		foreach (Lock doorLock in doorLocks) {
			if (doorLock.lockIndex == index) {
				return doorLock;
			}
		}
		return null;
	}

	Vector3 TileObjectPosFromPosXY(int posX, int posY, bool topLevel) {

		if (!topLevel) {
			posY = levelHeight - 1 - posY;
		}
		return new Vector3(posX - levelWidth / 2.0f + 0.5f, 0, posY - levelHeight / 2.0f + 0.5f);
	}

	TileType TileAtPos(int posX, int posY, bool topLevel) {
		
		return topLevel ? topTiles [posX, posY] : bottomTiles [posX, posY];
	}

	bool IsOnBoard(int posX, int posY) {

		return posX >= 0 && posX < levelWidth && posY >= 0 && posY < levelHeight;
	}

	Key KeyAtPos(int posX, int posY, bool topLevel) {

		foreach (Key key in keys) {
			if (key.forTopLevel == topLevel && key.posX == posX && key.posY == posY) {
				return key;
			}
		}
		return null;
	}
	Door DoorAtPos(int posX, int posY, bool topLevel) {
		
		foreach (Door door in doors) {
			if (door.forTopLevel == topLevel && door.posX == posX && door.posY == posY) {
				return door;
			}
		}
		return null;
	}

	bool IsKeyMovableByPlayer(Key key) {

		int nextPosX = playerPosX + (key.posX - playerPosX) * 2;
		int nextPosY = playerPosY + (key.posY - playerPosY) * 2;
		if (IsOnBoard(nextPosX, nextPosY) && KeyAtPos(nextPosX, nextPosY, playerOnTop) == null) {
			TileType tileType = TileAtPos(nextPosX, nextPosY, playerOnTop);
			if (tileType == TileType.Door) {
				Door door = DoorAtPos(nextPosX, nextPosY, playerOnTop);
				if (door.IsDoorLocked() || door.doorKey.keyIndex == key.keyIndex) {
					return false;
				}
			}
			return !(tileType == TileType.Gravel || tileType == TileType.Stone);
		}
		return false;
	}

	void CheckShadowGravelOrTrapOnPos(int posX, int posY) {

		GameObject gravel = null, shadowGravel = null;
		GameObject trap = null;
		if (playerOnTop) {
			if (bottomTiles[posX, posY] == TileType.Gravel) {
				shadowGravel = topGravel[posX, posY];
				topGravel[posX, posY] = null;
				gravel = bottomGravel[posX, posY];
				bottomGravel[posX, posY] = null;
				topTiles[posX, posY] = TileType.Empty;
				bottomTiles[posX, posY] = TileType.Empty;
			}
			else if (topTiles[posX, posY] == TileType.Trap) {
				trap = topGravel[posX, posY];
				topGravel[posX, posY] = null;
				topTiles[posX, posY] = TileType.Empty;
				bottomTiles[posX, posY] = TileType.Stone;
			}
		}
		else {
			if (topTiles[posX, posY] == TileType.Gravel) {
				shadowGravel = bottomGravel[posX, posY];
				bottomGravel[posX, posY] = null;
				gravel = topGravel[posX, posY];
				topGravel[posX, posY] = null;
				bottomTiles[posX, posY] = TileType.Empty;
				topTiles[posX, posY] = TileType.Empty;
			}
			else if (bottomTiles[posX, posY] == TileType.Trap) {
				trap = bottomGravel[posX, posY];
				bottomGravel[posX, posY] = null;
				bottomTiles[posX, posY] = TileType.Empty;
				topTiles[posX, posY] = TileType.Stone;
			}
		}
		
		if (trap) {
			Destroy(trap);
			if (!KeyAtPos(posX, posY, !playerOnTop)) {
				GameObject tileObject = (GameObject)Instantiate(stonePrefab);
				Transform levelTransform = !playerOnTop ? topLevel.transform : bottomLevel.transform;
				tileObject.transform.rotation = levelTransform.rotation;
				tileObject.transform.parent = levelTransform;
				tileObject.transform.localPosition = TileObjectPosFromPosXY(posX, posY, !playerOnTop);
			}
		}
		else if (gravel) {
			Destroy(gravel);
			Destroy(shadowGravel);
		}
	}

	public void ShowLevel(bool show) {

		gameObject.SetActive (show);
	}

	public void FinishLevel() {

		int levelIndex = 0;
		if (PlayerPrefs.HasKey("level")) {
			levelIndex = PlayerPrefs.GetInt("level");
		}
		levelIndex++;
		if (levelIndex >= LevelData.LevelCount()) {
			GUIManger.Instance.ShowMenu();
			return;
		}
		Debug.Log ("level" + levelIndex);
		PlayerPrefs.SetInt("level", levelIndex);

		int levelProgress = 0;
		if (PlayerPrefs.HasKey("progress")) {
			levelProgress = PlayerPrefs.GetInt("progress");
		}
		if (levelProgress < levelIndex) {
			PlayerPrefs.SetInt("progress", levelIndex);
		}
		CreateLevel();
	}

	public void RestartLevel() {

		CreateLevel ();
	}

	public bool CanMovePlayerWithDir(int dirX, int dirY) {

		int nextPosX = playerPosX + dirX;
		int nextPosY = playerPosY + dirY;
		if (!IsOnBoard(nextPosX, nextPosY)) {
			return false;
		}

		Key key = KeyAtPos (nextPosX, nextPosY, playerOnTop);
		if (key) {
			return IsKeyMovableByPlayer(key);
		}

		TileType tileType;
		if (playerOnTop) {
			tileType = topTiles[nextPosX, nextPosY];
		}
		else {
			tileType = bottomTiles[nextPosX, nextPosY];
		}

		if (tileType == TileType.Door) {
			Door door = DoorAtPos(nextPosX, nextPosY, playerOnTop);
			if (door.IsDoorLocked()) {
				return false;
			}
		}
		return !(tileType == TileType.Gravel || tileType == TileType.Stone);
	}

	public void MovePlayerWithDir(int dirX, int dirY) {
		
		int nextPosX = playerPosX + dirX;
		int nextPosY = playerPosY + dirY;
		Key key = KeyAtPos (nextPosX, nextPosY, playerOnTop);
		if (key && IsKeyMovableByPlayer(key)) {
			key.posX += dirX;
			key.posY += dirY;
			key.GetComponent<BoardMovementControl> ().Move (dirX, playerOnTop ? dirY : -dirY);
			CheckShadowGravelOrTrapOnPos(key.posX, key.posY);
		}

		playerPosX += dirX;
		playerPosY += dirY;
		player.GetComponent<BoardMovementControl> ().Move (dirX, dirY);
		CheckShadowGravelOrTrapOnPos(playerPosX, playerPosY);

		TileType tileType;
		if (playerOnTop) {
			tileType = topTiles[nextPosX, nextPosY];
		}
		else {
			tileType = bottomTiles[nextPosX, nextPosY];
		}
		
		if (tileType == TileType.Hole) {
			StartCoroutine ("SwitchPlayerSideAfterDelayCoroutine");
		}
		else if (tileType == TileType.Finish) {
			StartCoroutine ("FinishLevelAfterDelayCoroutine");
		}

		//Debug.Log (playerPosX + " " + playerPosY + " ");
	}

	public void SwitchPlayerSide() {

		playerOnTop = !playerOnTop;
		PlayerControl playerControl = player.GetComponent<PlayerControl> ();
		playerControl.topPlayer.renderer.enabled = playerOnTop;
		playerControl.bottomPlayer.renderer.enabled = !playerOnTop;
		playerControl.topShadowPlayer.renderer.enabled = !playerOnTop;
		playerControl.bottomShadowPlayer.renderer.enabled = playerOnTop;
		playerControl.waitForNoKeys = true;
		levelBoard.GetComponent<BoardFlipControl> ().Flip ();
	}

	public IEnumerator FinishLevelAfterDelayCoroutine() {
		
		yield return new WaitForSeconds(0.15f);
		FinishLevel ();
	}

	public IEnumerator SwitchPlayerSideAfterDelayCoroutine() {
		
		yield return new WaitForSeconds(0.15f);
		SwitchPlayerSide ();
	}
}
