﻿using UnityEngine;
using System.Collections;

public class PlayerControl : MonoBehaviour {

	public GameObject levelBoard;
	public GameObject player;
	public GameObject topPlayer;
	public GameObject topShadowPlayer;
	public GameObject bottomPlayer;
	public GameObject bottomShadowPlayer;
	public bool waitForNoKeys = true;

	private LevelModel levelModel;
	private BoardMovementControl playerMovementControl;
	private BoardFlipControl boardFlipControl;
	private float movementDelayTimer = 0;
	private KeyCode lastMoveKeyCode;


	void Awake() {

	}

	void OnDestroy() {

	}

	void Start () {

		levelModel = LevelModel.Instance;
		playerMovementControl = GetComponent<BoardMovementControl> ();
		boardFlipControl = levelBoard.GetComponent<BoardFlipControl> ();
	}
	
	void Update () {

		if (movementDelayTimer > 0) {
			movementDelayTimer -= Time.deltaTime;
		}
	
		if (Input.GetKeyDown (KeyCode.Space)) {
			boardFlipControl.Flip();
		}
		if (Input.GetKeyDown (KeyCode.Return)) {
		//	levelModel.SwitchPlayerSide();
		}
		if (Input.GetKeyDown(KeyCode.R)) {
			levelModel.RestartLevel();
		}
		if (Input.GetKeyDown(KeyCode.F)) {
		//	levelModel.FinishLevel();
		}
		HandleArrowKeys ();

		if (Input.GetKeyDown(KeyCode.Escape)) {
			GUIManger.Instance.ShowMenu();
		}
	}

	void HandleArrowKeys() {

		if (!Input.GetKey (KeyCode.LeftArrow) && !Input.GetKey (KeyCode.RightArrow) && !Input.GetKey (KeyCode.UpArrow) && !Input.GetKey (KeyCode.DownArrow)) {
			waitForNoKeys = false;
			movementDelayTimer = 0;
		}
		if (waitForNoKeys) {
			return;
		}
		if (!playerMovementControl.movementActive && !boardFlipControl.flipActive && IsFlippedOnPlayerSide()) {
			if (Input.GetKey (KeyCode.LeftArrow) && levelModel.CanMovePlayerWithDir(-1, 0)) {
				if (lastMoveKeyCode != KeyCode.LeftArrow || movementDelayTimer <= 0) {
					lastMoveKeyCode = KeyCode.LeftArrow;
					movementDelayTimer = 0.4f;
					levelModel.MovePlayerWithDir(-1, 0);
					return;
				}
			}
			if (Input.GetKey (KeyCode.RightArrow) && levelModel.CanMovePlayerWithDir(1, 0)) {
				if (lastMoveKeyCode != KeyCode.RightArrow || movementDelayTimer <= 0) {
					lastMoveKeyCode = KeyCode.RightArrow;
					movementDelayTimer = 0.4f;
					levelModel.MovePlayerWithDir(1, 0);
					return;
				}
			}
			if (Input.GetKey (KeyCode.UpArrow) && levelModel.CanMovePlayerWithDir(0, levelModel.playerOnTop ? 1 : -1)) {
				if (lastMoveKeyCode != KeyCode.UpArrow || movementDelayTimer <= 0) {
					lastMoveKeyCode = KeyCode.UpArrow;
					movementDelayTimer = 0.4f;
					levelModel.MovePlayerWithDir(0, levelModel.playerOnTop ? 1 : -1);
					return;
				}
			}
			if (Input.GetKey (KeyCode.DownArrow) && levelModel.CanMovePlayerWithDir(0, levelModel.playerOnTop ? -1 : 1)) {
				if (lastMoveKeyCode != KeyCode.DownArrow || movementDelayTimer <= 0) {
					lastMoveKeyCode = KeyCode.DownArrow;
					movementDelayTimer = 0.4f;
					levelModel.MovePlayerWithDir(0, levelModel.playerOnTop ? -1 : 1);
					return;
				}
			}
		}
	}

	bool IsFlippedOnPlayerSide() {

		return (levelModel.playerOnTop && boardFlipControl.showingTop) || (!levelModel.playerOnTop && !boardFlipControl.showingTop);
	}
}
